@echo off
REM MHA Toolbox Pro - Quick Launcher
REM Version 2.0.0

echo.
echo =====================================
echo   MHA Toolbox Pro - Version 2.0.0
echo   95+ Algorithms Platform
echo =====================================
echo.
echo Starting web interface...
echo.
echo The interface will open at: http://localhost:8501
echo.
echo Press Ctrl+C to stop the server
echo.

streamlit run mha_toolbox_pro_ultimate.py

pause
